@@ -1,10 +1,12 @@
 import binascii
 import enum
+import gzip
 import io
 import json
 import logging
 import mimetypes
 import os
+import warnings
 from typing import Any
 from typing import AnyStr
 from typing import Dict
@@ -19,8 +21,9 @@ from urllib.parse import quote
 from gcloud.aio.auth import AioSession  # pylint: disable=no-name-in-module
 from gcloud.aio.auth import BUILD_GCLOUD_REST  # pylint: disable=no-name-in-module
 from gcloud.aio.auth import Token  # pylint: disable=no-name-in-module
-from gcloud.aio.storage.bucket import Bucket
-from gcloud.aio.storage.constants import DEFAULT_TIMEOUT
+
+from .bucket import Bucket
+from .constants import DEFAULT_TIMEOUT
 
 # Selectively load libraries based on the package
 if BUILD_GCLOUD_REST:
@@ -50,7 +53,11 @@ def init_api_root(api_root: Optional[str]) -> Tuple[bool, str]:
 
     host = os.environ.get('STORAGE_EMULATOR_HOST')
     if host:
-        return True, f'http://{host}'
+        if not host.startswith('http'):
+            warnings.warn('STORAGE_EMULATOR_HOST must include http:// prefix',
+                          DeprecationWarning)
+            host = f'http://{host}'
+        return True, host
 
     return False, 'https://www.googleapis.com'
 
@@ -73,7 +80,7 @@ def encode_multipart_formdata(
     """
     body: List[bytes] = []
     for headers, data in fields:
-        body.append(f'--{boundary}\r\n'.encode('utf-8'))
+        body.append(f'--{boundary}\r\n'.encode())
 
         # The below is from RequestFields.render_headers()
         # Since we only use Content-Type, we could simplify the below to a
@@ -84,19 +91,19 @@ def encode_multipart_formdata(
         ]:
             value = headers.pop(field, None)
             if value:
-                body.append(f'{field}: {value}\r\n'.encode('utf-8'))
+                body.append(f'{field}: {value}\r\n'.encode())
         for field, value in headers.items():
             # N.B. potential bug copied from urllib3 code; zero values should
             # be sent! Keeping it for now, since Google libs use urllib3 for
             # their examples.
             if value:
-                body.append(f'{field}: {value}\r\n'.encode('utf-8'))
+                body.append(f'{field}: {value}\r\n'.encode())
 
         body.append(b'\r\n')
         body.append(data)
         body.append(b'\r\n')
 
-    body.append(f'--{boundary}--\r\n'.encode('utf-8'))
+    body.append(f'--{boundary}--\r\n'.encode())
 
     # N.B. 'multipart/form-data' in upstream, but Google wants 'related'
     content_type = f'multipart/related; boundary={boundary}'
@@ -111,7 +118,8 @@ class UploadType(enum.Enum):
 
 
 class StreamResponse:
-    """This class provides an abstraction between the slightly different
+    """
+    This class provides an abstraction between the slightly different
     recommended streaming implementations between requests and aiohttp.
     """
 
@@ -172,10 +180,41 @@ class Storage:
             'Authorization': f'Bearer {token}',
         }
 
+    # This method makes the following API call:
+    # https://cloud.google.com/storage/docs/json_api/v1/buckets/list
+    async def list_buckets(
+        self, project: str, *,
+        params: Optional[Dict[str, str]] = None,
+        headers: Optional[Dict[str, Any]] = None,
+        session: Optional[Session] = None,
+        timeout: int = DEFAULT_TIMEOUT,
+    ) -> List[Bucket]:
+        url = f'{self._api_root_read}?project={project}'
+        headers = headers or {}
+        headers.update(await self._headers())
+        params = params or {}
+        if not params.get('pageToken'):
+            params['pageToken'] = ''
+        s = AioSession(session) if session else self.session
+        buckets = []
+
+        while True:
+            resp = await s.get(url, headers=headers,
+                               params=params or {},
+                               timeout=timeout)
+
+            content: Dict[str, Any] = await resp.json(content_type=None)
+            for item in content.get('items', []):
+                buckets.append(Bucket(self, item['id']))
+
+            params['pageToken'] = content.get('nextPageToken', '')
+            if not params['pageToken']:
+                break
+        return buckets
+
     def get_bucket(self, bucket_name: str) -> Bucket:
         return Bucket(self, bucket_name)
 
-    # pylint: disable=too-many-locals
     async def copy(
         self, bucket: str, object_name: str,
         destination_bucket: str, *, new_name: Optional[str] = None,
@@ -183,25 +222,26 @@ class Storage:
         params: Optional[Dict[str, str]] = None,
         headers: Optional[Dict[str, str]] = None,
         timeout: int = DEFAULT_TIMEOUT,
-        session: Optional[Session] = None
+        session: Optional[Session] = None,
     ) -> Dict[str, Any]:
-
         """
-        When files are too large, multiple calls to `rewriteTo` are made. We
-        refer to the same copy job by using the `rewriteToken` from the
-        previous return payload in subsequent `rewriteTo` calls.
+        When files are too large, multiple calls to ``rewriteTo`` are made. We
+        refer to the same copy job by using the ``rewriteToken`` from the
+        previous return payload in subsequent ``rewriteTo`` calls.
 
-        Using the `rewriteTo` GCS API is preferred in part because it is able
-        to make multiple calls to fully copy an object whereas the `copyTo` GCS
-        API only calls `rewriteTo` once under the hood, and thus may fail if
-        files are large.
+        Using the ``rewriteTo`` GCS API is preferred in part because it is able
+        to make multiple calls to fully copy an object whereas the ``copyTo``
+        GCS API only calls ``rewriteTo`` once under the hood, and thus may fail
+        if files are large.
 
         In the rare case you need to resume a copy operation, include the
-        `rewriteToken` in the `params` dictionary. Once you begin a multi-part
-        copy operation, you then have 1 week to complete the copy job.
+        ``rewriteToken`` in the ``params`` dictionary. Once you begin a
+        multi-part copy operation, you then have 1 week to complete the copy
+        job.
 
-        https://cloud.google.com/storage/docs/json_api/v1/objects/rewrite
+        See https://cloud.google.com/storage/docs/json_api/v1/objects/rewrite
         """
+        # pylint: disable=too-many-locals
         if not new_name:
             new_name = object_name
 
@@ -249,7 +289,7 @@ class Storage:
             params['rewriteToken'] = data['rewriteToken']
             resp = await s.post(
                 url, headers=headers, params=params,
-                timeout=timeout,
+                timeout=timeout, data=metadata_,
             )
             data = await resp.json(content_type=None)
 
@@ -260,7 +300,7 @@ class Storage:
         timeout: int = DEFAULT_TIMEOUT,
         params: Optional[Dict[str, str]] = None,
         headers: Optional[Dict[str, str]] = None,
-        session: Optional[Session] = None
+        session: Optional[Session] = None,
     ) -> str:
         # https://cloud.google.com/storage/docs/request-endpoints#encoding
         encoded_object_name = quote(object_name, safe='')
@@ -285,7 +325,7 @@ class Storage:
         self, bucket: str, object_name: str, *,
         headers: Optional[Dict[str, Any]] = None,
         timeout: int = DEFAULT_TIMEOUT,
-        session: Optional[Session] = None
+        session: Optional[Session] = None,
     ) -> bytes:
         return await self._download(
             bucket, object_name, headers=headers,
@@ -295,7 +335,7 @@ class Storage:
 
     async def download_to_filename(
         self, bucket: str, object_name: str,
-        filename: str, **kwargs: Any
+        filename: str, **kwargs: Any,
     ) -> None:
         async with file_open(  # type: ignore[attr-defined]
                 filename,
@@ -309,11 +349,12 @@ class Storage:
         self, bucket: str, object_name: str, *,
         headers: Optional[Dict[str, Any]] = None,
         session: Optional[Session] = None,
-        timeout: int = DEFAULT_TIMEOUT
+        timeout: int = DEFAULT_TIMEOUT,
     ) -> Dict[str, Any]:
         data = await self._download(
             bucket, object_name, headers=headers,
-            timeout=timeout, session=session,
+            timeout=timeout, params={'alt': 'json'},
+            session=session,
         )
         metadata: Dict[str, Any] = json.loads(data.decode())
         return metadata
@@ -322,20 +363,19 @@ class Storage:
         self, bucket: str, object_name: str, *,
         headers: Optional[Dict[str, Any]] = None,
         timeout: int = DEFAULT_TIMEOUT,
-        session: Optional[Session] = None
+        session: Optional[Session] = None,
     ) -> StreamResponse:
-        """Download a GCS object in a buffered stream.
+        """
+        Download a GCS object in a buffered stream.
 
         Args:
-            bucket (str): The bucket from which to download.
-            object_name (str): The object within the bucket to download.
-            headers (Optional[Dict[str, Any]], optional): Custom header values
-                for the request, such as range. Defaults to None.
-            timeout (int, optional): Timeout, in seconds, for the request. Note
-                that with this function, this is the time to the beginning of
-                the response data (TTFB). Defaults to 10.
-            session (Optional[Session], optional): A specific session to
-                (re)use. Defaults to None.
+            bucket: The bucket from which to download.
+            object_name: The object within the bucket to download.
+            headers: Custom header values for the request, such as range.
+            timeout: Timeout, in seconds, for the request. Note that with this
+                function, this is the time to the beginning of the response
+                data (TTFB).
+            session: A specific session to (re)use.
 
         Returns:
             StreamResponse: A object encapsulating the stream, similar to
@@ -353,7 +393,7 @@ class Storage:
         params: Optional[Dict[str, str]] = None,
         headers: Optional[Dict[str, Any]] = None,
         session: Optional[Session] = None,
-        timeout: int = DEFAULT_TIMEOUT
+        timeout: int = DEFAULT_TIMEOUT,
     ) -> Dict[str, Any]:
         url = f'{self._api_root_read}/{bucket}/o'
         headers = headers or {}
@@ -377,12 +417,20 @@ class Storage:
         metadata: Optional[Dict[str, Any]] = None,
         session: Optional[Session] = None,
         force_resumable_upload: Optional[bool] = None,
-        timeout: int = 30
+        zipped: bool = False,
+        timeout: int = 30,
     ) -> Dict[str, Any]:
         url = f'{self._api_root_write}/{bucket}/o'
-
         stream = self._preprocess_data(file_data)
 
+        parameters = parameters or {}
+        if zipped:
+            parameters['contentEncoding'] = 'gzip'
+            # Here we load the file-like object data into memory in chunks and
+            # re-write it compressed. This is implemented like this so we don't
+            # load the whole file into memory at once.
+            stream = self._compress_file_in_chunks(input_stream=stream)
+
         if BUILD_GCLOUD_REST and isinstance(stream, io.StringIO):
             # HACK: `requests` library does not accept `str` as `data` in `put`
             # HTTP request.
@@ -393,8 +441,6 @@ class Storage:
         # mime detection method same as in aiohttp 3.4.4
         content_type = content_type or mimetypes.guess_type(object_name)[0]
 
-        parameters = parameters or {}
-
         headers = headers or {}
         headers.update(await self._headers())
         headers.update({
@@ -429,17 +475,14 @@ class Storage:
     async def upload_from_filename(
         self, bucket: str, object_name: str,
         filename: str,
-        **kwargs: Any
+        **kwargs: Any,
     ) -> Dict[str, Any]:
         async with file_open(  # type: ignore[attr-defined]
                 filename,
                 mode='rb',
         ) as file_object:
             contents = await file_object.read()
-            return await self.upload(
-                bucket, object_name, contents,
-                **kwargs
-            )
+            return await self.upload(bucket, object_name, contents, **kwargs)
 
     @staticmethod
     def _get_stream_len(stream: IO[AnyStr]) -> int:
@@ -463,6 +506,34 @@ class Storage:
 
         raise TypeError(f'unsupported upload type: "{type(data)}"')
 
+    @staticmethod
+    def _compress_file_in_chunks(input_stream: IO[AnyStr],
+                                 chunk_size: int = 8192) -> IO[bytes]:
+        """
+        Reads the contents of input_stream and writes it gzip-compressed to
+        output_stream in chunks. The chunk size is 8Kb by default, which is a
+        standard filesystem block size.
+        """
+        compressed_stream = io.BytesIO()
+
+        with gzip.open(compressed_stream, 'wb') as gzipped_file:
+            chunk_bytes: bytes
+            while True:
+                chunk = input_stream.read(chunk_size)
+                if not chunk:
+                    break
+                if isinstance(chunk, str):
+                    chunk_bytes = chunk.encode('utf-8')
+                else:
+                    chunk_bytes = chunk
+
+                gzipped_file.write(chunk_bytes)
+
+        # After finishing writing, reset the buffer position so it can be read
+        compressed_stream.seek(0)
+
+        return compressed_stream
+
     @staticmethod
     def _decide_upload_type(
         force_resumable_upload: Optional[bool],
@@ -510,7 +581,7 @@ class Storage:
         params: Optional[Dict[str, str]] = None,
         headers: Optional[Dict[str, str]] = None,
         timeout: int = DEFAULT_TIMEOUT,
-        session: Optional[Session] = None
+        session: Optional[Session] = None,
     ) -> bytes:
         # https://cloud.google.com/storage/docs/request-endpoints#encoding
         encoded_object_name = quote(object_name, safe='')
@@ -518,19 +589,35 @@ class Storage:
         headers = headers or {}
         headers.update(await self._headers())
 
-        s = AioSession(session) if session else self.session
-        response = await s.get(
-            url, headers=headers, params=params or {},
-            timeout=timeout,
-        )
+        # aiohttp and requests automatically decompress the body if this
+        # argument is not passed. We assume that if the Accept-Encoding header
+        # is present, then the client will handle the decompression
+        auto_decompress = 'accept-encoding' not in {k.lower() for k in headers}
 
-        # N.B. the GCS API sometimes returns 'application/octet-stream' when a
-        # string was uploaded. To avoid potential weirdness, always return a
-        # bytes object.
-        try:
-            data: bytes = await response.read()
-        except (AttributeError, TypeError):
-            data = response.content  # type: ignore[assignment]
+        s = AioSession(session) if session else self.session
+
+        data: bytes
+        if not auto_decompress and BUILD_GCLOUD_REST:
+            # Requests lib has a different way of reading compressed data. We
+            # must pass the stream=True argument and read the response using
+            # the 'raw' property.
+            response = await s.get(
+                url, headers=headers, params=params or {},
+                timeout=timeout, stream=True,
+            )
+            data = response.raw.read()  # type: ignore[attr-defined]
+        else:
+            response = await s.get(
+                url, headers=headers, params=params or {},
+                timeout=timeout, auto_decompress=auto_decompress,
+            )
+            # N.B. the GCS API sometimes returns 'application/octet-stream'
+            # when a string was uploaded. To avoid potential weirdness, always
+            # return a bytes object.
+            try:
+                data = await response.read()
+            except (AttributeError, TypeError):
+                data = response.content  # type: ignore[assignment]
 
         return data
 
@@ -539,7 +626,7 @@ class Storage:
         params: Optional[Dict[str, str]] = None,
         headers: Optional[Dict[str, str]] = None,
         timeout: int = DEFAULT_TIMEOUT,
-        session: Optional[Session] = None
+        session: Optional[Session] = None,
     ) -> StreamResponse:
         # https://cloud.google.com/storage/docs/request-endpoints#encoding
         encoded_object_name = quote(object_name, safe='')
@@ -570,7 +657,7 @@ class Storage:
         stream: IO[AnyStr], params: Dict[str, str],
         headers: Dict[str, str], *,
         session: Optional[Session] = None,
-        timeout: int = 30
+        timeout: int = 30,
     ) -> Dict[str, Any]:
         # https://cloud.google.com/storage/docs/json_api/v1/how-tos/simple-upload
         params['name'] = object_name
@@ -590,7 +677,7 @@ class Storage:
         headers: Dict[str, str],
         metadata: Dict[str, Any], *,
         session: Optional[Session] = None,
-        timeout: int = 30
+        timeout: int = 30,
     ) -> Dict[str, Any]:
         # https://cloud.google.com/storage/docs/json_api/v1/how-tos/multipart-upload
         params['uploadType'] = 'multipart'
@@ -645,7 +732,7 @@ class Storage:
         headers: Dict[str, str], *,
         metadata: Optional[Dict[str, Any]] = None,
         session: Optional[Session] = None,
-        timeout: int = 30
+        timeout: int = 30,
     ) -> Dict[str, Any]:
         # https://cloud.google.com/storage/docs/json_api/v1/how-tos/resumable-upload
         session_uri = await self._initiate_upload(
@@ -663,7 +750,7 @@ class Storage:
         params: Dict[str, str], headers: Dict[str, str],
         *, metadata: Optional[Dict[str, Any]] = None,
         timeout: int = DEFAULT_TIMEOUT,
-        session: Optional[Session] = None
+        session: Optional[Session] = None,
     ) -> str:
         params['uploadType'] = 'resumable'
 
@@ -701,14 +788,15 @@ class Storage:
         self, session_uri: str, stream: IO[AnyStr],
         headers: Dict[str, str], *, retries: int = 5,
         session: Optional[Session] = None,
-        timeout: int = 30
+        timeout: int = 30,
     ) -> Dict[str, Any]:
         s = AioSession(session) if session else self.session
 
         original_close = stream.close
         original_position = stream.tell()
         # Prevent the stream being closed if put operation fails
-        stream.close = lambda: None  # type: ignore[assignment]
+        stream.close = lambda: None  # type: ignore[method-assign]
+        resp = None
         try:
             for tries in range(retries):
                 try:
@@ -717,7 +805,10 @@ class Storage:
                         data=stream, timeout=timeout,
                     )
                 except ResponseError:
-                    headers.update({'Content-Range': '*/*'})
+                    if tries == retries - 1:
+                        raise
+
+                    headers.update({'Content-Range': 'bytes */*'})
                     stream.seek(original_position)
 
                     await sleep(  # type: ignore[func-returns-value]
@@ -728,6 +819,9 @@ class Storage:
         finally:
             original_close()
 
+        if resp is None:
+            raise ValueError('null response, this should not happen')
+
         data: Dict[str, Any] = await resp.json(content_type=None)
         return data
 
@@ -736,7 +830,7 @@ class Storage:
             *, params: Optional[Dict[str, str]] = None,
             headers: Optional[Dict[str, str]] = None,
             session: Optional[Session] = None,
-            timeout: int = DEFAULT_TIMEOUT
+            timeout: int = DEFAULT_TIMEOUT,
     ) -> Dict[str, Any]:
         # https://cloud.google.com/storage/docs/json_api/v1/objects/patch
         encoded_object_name = quote(object_name, safe='')
@@ -760,7 +854,7 @@ class Storage:
         params: Optional[Dict[str, str]] = None,
         headers: Optional[Dict[str, str]] = None,
         session: Optional[Session] = None,
-        timeout: int = DEFAULT_TIMEOUT
+        timeout: int = DEFAULT_TIMEOUT,
     ) -> Dict[str, Any]:
         url = f'{self._api_root_read}/{bucket}'
         headers = headers or {}
